from datetime import datetime

from eq3btsmart.adapter.base_adapter import BaseAdapter


class Eq3Time(BaseAdapter[datetime, bytes]):
    """Adapter to encode and decode time data."""

    @classmethod
    def _encode(cls, value: datetime) -> bytes:
        return bytes(
            [
                value.year % 100,
                value.month,
                value.day,
                value.hour,
                value.minute,
                value.second,
            ]
        )

    @classmethod
    def _decode(cls, value: bytes) -> datetime:
        (year, month, day, hour, minute, second) = value
        year += 2000

        return datetime(
            year=year, month=month, day=day, hour=hour, minute=minute, second=second
        )



from datetime import time

from eq3btsmart.adapter.base_adapter import BaseAdapter


class Eq3ScheduleTime(BaseAdapter[time, int]):
    """Adapter to encode and decode schedule time data."""

    @classmethod
    def _encode(cls, value: time) -> int:
        return int((value.hour * 60 + value.minute) / 10)

    @classmethod
    def _decode(cls, value: int) -> time:
        hour, minute = divmod(value * 10, 60)
        return time(hour=hour, minute=minute)



from eq3btsmart.adapter.base_adapter import BaseAdapter
from eq3btsmart.const import EQ3BT_MAX_OFFSET, EQ3BT_MIN_OFFSET


class Eq3TemperatureOffset(BaseAdapter[float, int]):
    """Adapter to encode and decode temperature offset data."""

    @classmethod
    def _encode(cls, value: float) -> int:
        if value < EQ3BT_MIN_OFFSET or value > EQ3BT_MAX_OFFSET:
            raise ValueError(
                f"Temperature {value} out of range [{EQ3BT_MIN_OFFSET}, {EQ3BT_MAX_OFFSET}]"
            )

        return int((value + 3.5) / 0.5)

    @classmethod
    def _decode(cls, value: int) -> float:
        return float(value * 0.5 - 3.5)



from dataclasses import dataclass

from eq3btsmart.adapter.eq3_schedule_time import Eq3ScheduleTime
from eq3btsmart.adapter.eq3_temperature import Eq3Temperature


@dataclass
class ScheduleHour:
    target_temperature: Eq3Temperature
    next_change_at: Eq3ScheduleTime



from datetime import datetime, timedelta

from eq3btsmart.adapter.base_adapter import BaseAdapter


class Eq3AwayTime(BaseAdapter[datetime, bytes]):
    """Adapter to encode and decode away time data."""

    @classmethod
    def _encode(cls, value: datetime) -> bytes:
        value += timedelta(minutes=15)
        value -= timedelta(minutes=value.minute % 30)

        if value.year < 2000 or value.year > 2099:
            raise ValueError("Invalid year, possible [2000, 2099]")

        year = value.year - 2000
        hour = value.hour * 2
        if value.minute != 0:
            hour |= 0x01

        return bytes([value.day, year, hour, value.month])

    @classmethod
    def _decode(cls, value: bytes) -> datetime:
        if value == bytes([0x00, 0x00, 0x00, 0x00]):
            return datetime(year=2000, month=1, day=1, hour=0, minute=0)

        (day, year, hour_min, month) = value

        year += 2000

        min = 0
        if hour_min & 0x01:
            min = 30
        hour = int(hour_min / 2)

        return datetime(year=year, month=month, day=day, hour=hour, minute=min)



import asyncio

from bleak import BleakClient

from eq3btsmart.const import MONITOR_INTERVAL


class ConnectionMonitor:
    def __init__(self, client: BleakClient):
        self._client = client
        self._run: bool = False

    async def run(self):
        self._run = True

        while self._run:
            await self._check_connection()

            await asyncio.sleep(MONITOR_INTERVAL)

    async def _check_connection(self):
        if self._run:
            try:
                if not self._client.is_connected:
                    await self._client.connect()
            except Exception:
                pass

    async def stop(self):
        self._run = False



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, Self, Type, TypeVar

from eq3btsmart.structures import Eq3Struct

StructType = TypeVar("StructType", bound=Eq3Struct)


@dataclass
class BaseModel(ABC, Generic[StructType]):
    @classmethod
    @abstractmethod
    def from_struct(cls, struct: StructType) -> Self:
        """Convert the structure to a model."""

    @classmethod
    @abstractmethod
    def struct_type(cls) -> Type[StructType]:
        """Convert the model to a structure."""

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        """Convert the data to a model."""

        return cls.from_struct(cls.struct_type().from_bytes(data))



from eq3btsmart.adapter.base_adapter import BaseAdapter
from eq3btsmart.const import (
    EQ3BT_OFF_TEMP,
    EQ3BT_ON_TEMP,
)


class Eq3Temperature(BaseAdapter[float, int]):
    """Adapter to encode and decode temperature data."""

    @classmethod
    def _encode(cls, value: float) -> int:
        if value < EQ3BT_OFF_TEMP or value > EQ3BT_ON_TEMP:
            raise ValueError(
                f"Temperature {value} out of range [{EQ3BT_OFF_TEMP}, {EQ3BT_ON_TEMP}]"
            )

        return int(value * 2)

    @classmethod
    def _decode(cls, value: int) -> float:
        return float(value / 2)



from abc import ABC, abstractmethod
from typing import Generic, Type, TypeVar

from construct import Adapter

EncodedType = TypeVar("EncodedType")
DecodedType = TypeVar("DecodedType")


class BaseAdapter(ABC, Generic[DecodedType, EncodedType]):
    """Base class for adapters."""

    def __init__(self, value: DecodedType):
        """Initialize the parser."""

        self._value: DecodedType = value

    @property
    def value(self) -> DecodedType:
        """Return the original value."""

        return self._value

    @classmethod
    def adapter(cls) -> Type[Adapter]:
        """Return the adapter."""

        class Eq3Adapter(Adapter):
            def _decode(
                self, obj: EncodedType, ctx, path
            ) -> BaseAdapter[DecodedType, EncodedType]:
                return cls(cls._decode(obj))

            def _encode(self, obj: BaseAdapter, ctx, path) -> EncodedType:
                return obj.encode()

        return Eq3Adapter

    def encode(self) -> EncodedType:
        """Return the bytes from the parser."""

        return self._encode(self.value)

    @classmethod
    def decode(cls, value: EncodedType) -> "BaseAdapter":
        """Return the parser from bytes."""

        return cls(cls._decode(value))

    @classmethod
    @abstractmethod
    def _encode(cls, value: DecodedType) -> EncodedType:
        """Return the encoded value."""

    @classmethod
    @abstractmethod
    def _decode(cls, value: EncodedType) -> DecodedType:
        """Return the decoded value."""

    def __eq__(self, __value: object) -> bool:
        return isinstance(__value, self.__class__) and self.value == __value.value

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.value})"



from dataclasses import dataclass
from typing import Self

from eq3btsmart.adapter.eq3_away_time import Eq3AwayTime
from eq3btsmart.adapter.eq3_temperature import Eq3Temperature
from eq3btsmart.const import EQ3BT_OFF_TEMP, EQ3BT_ON_TEMP, OperationMode
from eq3btsmart.models import BaseModel, Presets
from eq3btsmart.structures import StatusStruct


@dataclass
class Status(BaseModel[StatusStruct]):
    valve: int
    target_temperature: Eq3Temperature
    _operation_mode: OperationMode
    is_away: bool
    is_boost: bool
    is_dst: bool
    is_window_open: bool
    is_locked: bool
    is_low_battery: bool
    away_until: Eq3AwayTime | None = None
    presets: Presets | None = None

    @property
    def operation_mode(self) -> OperationMode:
        if self.target_temperature.value == EQ3BT_OFF_TEMP:
            return OperationMode.OFF

        if self.target_temperature.value == EQ3BT_ON_TEMP:
            return OperationMode.ON

        return self._operation_mode

    @property
    def valve_temperature(self) -> float:
        return (1 - self.valve / 100) * 2 + self.target_temperature.value - 2

    @classmethod
    def from_struct(cls, struct: StatusStruct) -> Self:
        return cls(
            valve=struct.valve,
            target_temperature=struct.target_temp,
            _operation_mode=OperationMode.MANUAL
            if struct.mode & struct.mode.MANUAL
            else OperationMode.AUTO,
            is_away=bool(struct.mode & struct.mode.AWAY),
            is_boost=bool(struct.mode & struct.mode.BOOST),
            is_dst=bool(struct.mode & struct.mode.DST),
            is_window_open=bool(struct.mode & struct.mode.WINDOW),
            is_locked=bool(struct.mode & struct.mode.LOCKED),
            is_low_battery=bool(struct.mode & struct.mode.LOW_BATTERY),
            away_until=struct.away,
            presets=Presets.from_struct(struct.presets) if struct.presets else None,
        )

    @classmethod
    def struct_type(cls) -> type[StatusStruct]:
        return StatusStruct



from dataclasses import dataclass
from typing import Self

from eq3btsmart.adapter.eq3_serial import Eq3Serial
from eq3btsmart.models.base_model import BaseModel
from eq3btsmart.structures import DeviceDataStruct


@dataclass
class DeviceData(BaseModel[DeviceDataStruct]):
    firmware_version: int
    device_serial: Eq3Serial

    @classmethod
    def from_struct(cls, struct: DeviceDataStruct) -> Self:
        return cls(
            firmware_version=struct.version,
            device_serial=struct.serial,
        )

    @classmethod
    def struct_type(cls) -> type[DeviceDataStruct]:
        return DeviceDataStruct



"""
Support for eq3 Bluetooth Smart thermostats.

All temperatures in Celsius.

To get the current state, update() has to be called for powersaving reasons.
"""

import asyncio
import logging
from datetime import datetime, timedelta
from typing import Callable

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.exc import BleakError
from construct_typed import DataclassStruct

from eq3btsmart.adapter.eq3_away_time import Eq3AwayTime
from eq3btsmart.adapter.eq3_duration import Eq3Duration
from eq3btsmart.adapter.eq3_temperature import Eq3Temperature
from eq3btsmart.adapter.eq3_temperature_offset import Eq3TemperatureOffset
from eq3btsmart.adapter.eq3_time import Eq3Time
from eq3btsmart.connection_monitor import ConnectionMonitor
from eq3btsmart.const import (
    DEFAULT_AWAY_HOURS,
    DEFAULT_AWAY_TEMP,
    EQ3BT_MAX_TEMP,
    EQ3BT_MIN_TEMP,
    EQ3BT_OFF_TEMP,
    EQ3BT_ON_TEMP,
    NOTIFY_CHARACTERISTIC_UUID,
    REQUEST_TIMEOUT,
    WRITE_CHARACTERISTIC_UUID,
    Command,
    Eq3Preset,
    OperationMode,
    WeekDay,
)
from eq3btsmart.exceptions import Eq3Exception
from eq3btsmart.models import DeviceData, Schedule, Status
from eq3btsmart.structures import (
    AwaySetCommand,
    BoostSetCommand,
    ComfortEcoConfigureCommand,
    ComfortSetCommand,
    EcoSetCommand,
    Eq3Command,
    Eq3Struct,
    IdGetCommand,
    InfoGetCommand,
    LockSetCommand,
    ModeSetCommand,
    OffsetConfigureCommand,
    ScheduleGetCommand,
    ScheduleHourStruct,
    ScheduleSetCommand,
    TemperatureSetCommand,
    WindowOpenConfigureCommand,
)
from eq3btsmart.thermostat_config import ThermostatConfig

_LOGGER = logging.getLogger(__name__)


class Thermostat:
    """Representation of a EQ3 Bluetooth Smart thermostat."""

    def __init__(
        self,
        thermostat_config: ThermostatConfig,
        ble_device: BLEDevice,
    ):
        """Initialize the thermostat."""

        self.thermostat_config = thermostat_config
        self.status: Status | None = None
        self.device_data: DeviceData | None = None
        self.schedule: Schedule = Schedule()
        self._on_update_callbacks: list[Callable] = []
        self._on_connection_callbacks: list[Callable] = []
        self._device = ble_device
        self._conn: BleakClient = BleakClient(
            ble_device,
            disconnected_callback=lambda client: self._on_connection_changed(False),
            timeout=REQUEST_TIMEOUT,
        )
        self._lock = asyncio.Lock()
        self._monitor = ConnectionMonitor(self._conn)

    @property
    def is_connected(self) -> bool:
        """Return true if connected to the thermostat."""

        return self._conn.is_connected

    @property
    def is_monitoring(self) -> bool:
        """Return true if monitoring the connection."""

        return self._monitor._run

    @property
    def is_busy(self) -> bool:
        """Return true if the thermostat is busy."""

        return self._lock.locked()

    def register_connection_callback(self, on_connect: Callable) -> None:
        """Register a callback function that will be called when a connection is established."""

        self._on_connection_callbacks.append(on_connect)

    def register_update_callback(self, on_update: Callable) -> None:
        """Register a callback function that will be called when an update is received."""

        self._on_update_callbacks.append(on_update)

    async def async_connect(self) -> None:
        """Connect to the thermostat."""

        await self._conn.connect()
        await self._conn.start_notify(NOTIFY_CHARACTERISTIC_UUID, self.on_notification)

        self._on_connection_changed(True)

        loop = asyncio.get_running_loop()
        loop.create_task(self._monitor.run())

    async def async_disconnect(self) -> None:
        """Shutdown the connection to the thermostat."""

        await self._monitor.stop()
        await self._conn.disconnect()

    async def async_get_id(self) -> None:
        """Query device identification information, e.g. the serial number."""

        await self._async_write_command(IdGetCommand())

    async def async_get_status(self) -> None:
        """Query the thermostat status."""

        eq3_time = Eq3Time(datetime.now())
        await self._async_write_command(InfoGetCommand(time=eq3_time))

    async def async_get_schedule(self) -> None:
        """Query the schedule."""

        for week_day in WeekDay:
            await self._async_write_command(ScheduleGetCommand(day=week_day))

    async def async_configure_window_open(
        self, temperature: float, duration: timedelta
    ) -> None:
        """Configures the window open behavior. The duration is specified in 5 minute increments."""

        eq3_temperature = Eq3Temperature(temperature)
        eq3_duration = Eq3Duration(duration)

        await self._async_write_command(
            WindowOpenConfigureCommand(
                window_open_temperature=eq3_temperature,
                window_open_time=eq3_duration,
            )
        )

    async def async_configure_presets(
        self,
        comfort_temperature: float | None = None,
        eco_temperature: float | None = None,
    ) -> None:
        """Set the thermostats preset temperatures comfort (sun) and eco (moon)."""

        if self.status is None:
            raise Eq3Exception("Status not set")

        if comfort_temperature is None and self.status.presets is not None:
            comfort_temperature = self.status.presets.comfort_temperature.value

        if eco_temperature is None and self.status.presets is not None:
            eco_temperature = self.status.presets.eco_temperature.value

        if comfort_temperature is None or eco_temperature is None:
            raise Eq3Exception("Comfort or eco temperature not set")

        eq3_comfort_temperature = Eq3Temperature(comfort_temperature)
        eq3_eco_temperature = Eq3Temperature(eco_temperature)

        await self._async_write_command(
            ComfortEcoConfigureCommand(
                comfort_temperature=eq3_comfort_temperature,
                eco_temperature=eq3_eco_temperature,
            )
        )

    async def async_configure_temperature_offset(
        self, temperature_offset: float
    ) -> None:
        """Sets the thermostat's temperature offset."""

        eq3_temperature_offset = Eq3TemperatureOffset(temperature_offset)
        await self._async_write_command(
            OffsetConfigureCommand(offset=eq3_temperature_offset)
        )

    async def async_set_mode(self, operation_mode: OperationMode) -> None:
        """Set new operation mode."""

        if self.status is None or self.status.target_temperature is None:
            raise Eq3Exception("Status not set")

        command: ModeSetCommand

        match operation_mode:
            case OperationMode.AUTO:
                command = ModeSetCommand(mode=OperationMode.AUTO)
            case OperationMode.MANUAL:
                temperature = max(
                    min(
                        self.status.target_temperature.value,
                        EQ3BT_MAX_TEMP,
                    ),
                    EQ3BT_MIN_TEMP,
                )
                command = ModeSetCommand(
                    mode=OperationMode.MANUAL | Eq3Temperature(temperature).encode()
                )
            case OperationMode.OFF:
                off_temperature = Eq3Temperature(EQ3BT_OFF_TEMP)
                command = ModeSetCommand(
                    mode=OperationMode.MANUAL | off_temperature.encode()
                )
            case OperationMode.ON:
                if self.status is not None and self.status.is_away:
                    await self.async_set_away(False)

                if self.status is not None and self.status.is_boost:
                    await self.async_set_boost(False)

                on_temperature = Eq3Temperature(EQ3BT_ON_TEMP)
                command = ModeSetCommand(
                    mode=OperationMode.MANUAL | on_temperature.encode()
                )

        await self._async_write_command(command)

    async def async_set_away(
        self,
        enable: bool,
        away_until: datetime | None = None,
        temperature: float | None = None,
    ) -> None:
        if not enable:
            return await self.async_set_mode(OperationMode.AUTO)

        if away_until is None:
            away_until = datetime.now() + timedelta(hours=DEFAULT_AWAY_HOURS)

        if temperature is None:
            temperature = DEFAULT_AWAY_TEMP

        eq3_away_until = Eq3AwayTime(away_until)
        eq3_temperature = Eq3Temperature(temperature)

        await self._async_write_command(
            AwaySetCommand(
                mode=OperationMode.AWAY | eq3_temperature.encode(),
                away_until=eq3_away_until,
            )
        )

    async def async_set_temperature(self, temperature: float) -> None:
        """Set new target temperature."""

        if temperature == EQ3BT_OFF_TEMP:
            return await self.async_set_mode(OperationMode.OFF)

        if temperature == EQ3BT_ON_TEMP:
            return await self.async_set_mode(OperationMode.ON)

        eq3_temperature = Eq3Temperature(temperature)
        await self._async_write_command(
            TemperatureSetCommand(temperature=eq3_temperature)
        )

    async def async_set_preset(self, preset: Eq3Preset):
        """Sets the thermostat to the given preset."""

        command: ComfortSetCommand | EcoSetCommand

        match preset:
            case Eq3Preset.COMFORT:
                command = ComfortSetCommand()
            case Eq3Preset.ECO:
                command = EcoSetCommand()

        if self.status is not None and self.status.is_away:
            await self.async_set_away(False)

        if self.status is not None and self.status.is_boost:
            await self.async_set_boost(False)

        await self._async_write_command(command)

    async def async_set_boost(self, enable: bool) -> None:
        """Sets boost mode."""

        await self._async_write_command(BoostSetCommand(enable=enable))

    async def async_set_locked(self, enable: bool) -> None:
        """Locks or unlocks the thermostat."""

        await self._async_write_command(LockSetCommand(enable=enable))

    async def async_delete_schedule(self, week_day: WeekDay | None = None) -> None:
        """Deletes the schedule."""

        if week_day is not None:
            await self._async_write_command(ScheduleSetCommand(day=week_day, hours=[]))
        else:
            for week_day in WeekDay:
                await self._async_write_command(
                    ScheduleSetCommand(day=week_day, hours=[])
                )

    async def async_set_schedule(self, schedule: Schedule) -> None:
        """Sets the schedule for the given day."""

        for schedule_day in schedule.schedule_days:
            command = ScheduleSetCommand(
                day=schedule_day.week_day,
                hours=[
                    ScheduleHourStruct(
                        target_temp=schedule_hour.target_temperature,
                        next_change_at=schedule_hour.next_change_at,
                    )
                    for schedule_hour in schedule_day.schedule_hours
                ],
            )

            await self._async_write_command(command)

    async def _async_write_command(self, command: Eq3Struct) -> None:
        """Write a EQ3 command to the thermostat."""

        if not self._conn.is_connected:
            raise Eq3Exception("Not connected")

        data = command.to_bytes()

        async with self._lock:
            try:
                await self._conn.write_gatt_char(WRITE_CHARACTERISTIC_UUID, data)
            except (BleakError, TimeoutError) as ex:
                raise Eq3Exception("Error during write") from ex

    def _on_connection_changed(self, connected: bool) -> None:
        """Handle connection changes."""

        for callback in self._on_connection_callbacks:
            callback(connected)

    def on_notification(self, handle: BleakGATTCharacteristic, data: bytes) -> None:
        """Handle Callback from a Bluetooth (GATT) request."""

        data_bytes = bytes(data)

        command = DataclassStruct(Eq3Command).parse(data_bytes)

        is_status_command = command.data[0] == 0x01

        match command.cmd:
            case Command.ID_RETURN:
                self.device_data = DeviceData.from_bytes(data_bytes)
            case Command.INFO_RETURN:
                if is_status_command:
                    self.status = Status.from_bytes(data_bytes)
            case Command.SCHEDULE_RETURN:
                schedule = Schedule.from_bytes(data_bytes)
                self.schedule.merge(schedule)

        for callback in self._on_update_callbacks:
            callback()



from dataclasses import dataclass, field
from typing import Self

from eq3btsmart.const import WeekDay
from eq3btsmart.models.base_model import BaseModel
from eq3btsmart.models.schedule_hour import ScheduleHour
from eq3btsmart.structures import ScheduleDayStruct


@dataclass
class ScheduleDay(BaseModel[ScheduleDayStruct]):
    week_day: WeekDay
    schedule_hours: list[ScheduleHour] = field(default_factory=list)

    @classmethod
    def from_struct(cls, struct: ScheduleDayStruct) -> Self:
        return cls(
            week_day=struct.day,
            schedule_hours=[
                ScheduleHour(
                    target_temperature=hour.target_temp,
                    next_change_at=hour.next_change_at,
                )
                for hour in struct.hours
            ],
        )

    @classmethod
    def struct_type(cls) -> type[ScheduleDayStruct]:
        return ScheduleDayStruct

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, ScheduleDay):
            return False

        if self.week_day != __value.week_day:
            return False

        if len(self.schedule_hours) != len(__value.schedule_hours):
            return False

        return all(
            hour == other_hour
            for hour, other_hour in zip(self.schedule_hours, __value.schedule_hours)
        )



from eq3btsmart.adapter.base_adapter import BaseAdapter


class Eq3Serial(BaseAdapter[str, bytes]):
    """Adapter to encode and decode serial id."""

    @classmethod
    def _encode(cls, value: str) -> bytes:
        return bytes(char + 0x30 for char in value.encode())

    @classmethod
    def _decode(cls, value: bytes) -> str:
        return bytes(char - 0x30 for char in value).decode()



from datetime import timedelta

from eq3btsmart.adapter.base_adapter import BaseAdapter


class Eq3Duration(BaseAdapter[timedelta, int]):
    """Adapter to encode and decode duration data."""

    @classmethod
    def _encode(cls, value: timedelta) -> int:
        if value.seconds < 0 or value.seconds > 3600.0:
            raise ValueError(
                "Window open time must be between 0 and 60 minutes "
                "in intervals of 5 minutes."
            )

        return int(value.seconds / 300.0)

    @classmethod
    def _decode(cls, value: int) -> timedelta:
        return timedelta(minutes=float(value * 5.0))
